# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Quotas for resources per project."""

import copy

from oslo_log import log as logging
from oslo_utils import importutils
from sqlalchemy import sql

import nova.conf
from nova import context as nova_context
from nova.db.api import api as api_db_api
from nova.db.api import models as api_models
from nova.db.main import api as main_db_api
from nova import exception
from nova.limit import local as local_limit
from nova.limit import placement as placement_limit
from nova import objects
from nova.scheduler.client import report
from nova import utils

LOG = logging.getLogger(__name__)
CONF = nova.conf.CONF
# Lazy-loaded on first access.
# Avoid constructing the KSA adapter and provider tree on every access.
PLACEMENT_CLIENT = None
# If user_id and queued_for_delete are populated for a project, cache the
# result to avoid doing unnecessary EXISTS database queries.
UID_QFD_POPULATED_CACHE_BY_PROJECT = set()
# For the server group members check, we do not scope to a project, so if all
# user_id and queued_for_delete are populated for all projects, cache the
# result to avoid doing unnecessary EXISTS database queries.
UID_QFD_POPULATED_CACHE_ALL = False


class DbQuotaDriver(object):
    """Driver to perform necessary checks to enforce quotas and obtain
    quota information.  The default driver utilizes the local
    database.
    """
    UNLIMITED_VALUE = -1

    def get_reserved(self):
        # Since we stopped reserving the DB, we just return 0
        return 0

    def get_defaults(self, context, resources):
        """Given a list of resources, retrieve the default quotas.
        Use the class quotas named `_DEFAULT_QUOTA_NAME` as default quotas,
        if it exists.

        :param context: The request context, for access checks.
        :param resources: A dictionary of the registered resources.
        """

        quotas = {}
        default_quotas = objects.Quotas.get_default_class(context)
        for resource in resources.values():
            # resource.default returns the config options. So if there's not
            # an entry for the resource in the default class, it uses the
            # config option.
            quotas[resource.name] = default_quotas.get(resource.name,
                                                       resource.default)

        return quotas

    def get_class_quotas(self, context, resources, quota_class):
        """Given a list of resources, retrieve the quotas for the given
        quota class.

        :param context: The request context, for access checks.
        :param resources: A dictionary of the registered resources.
        :param quota_class: The name of the quota class to return
                            quotas for.
        """

        quotas = {}
        class_quotas = objects.Quotas.get_all_class_by_name(context,
                                                            quota_class)
        for resource in resources.values():
            quotas[resource.name] = class_quotas.get(resource.name,
                                                     resource.default)

        return quotas

    def _process_quotas(self, context, resources, project_id, quotas,
                        quota_class=None, usages=None,
                        remains=False):
        modified_quotas = {}
        # Get the quotas for the appropriate class.  If the project ID
        # matches the one in the context, we use the quota_class from
        # the context, otherwise, we use the provided quota_class (if
        # any)
        if project_id == context.project_id:
            quota_class = context.quota_class
        if quota_class:
            class_quotas = objects.Quotas.get_all_class_by_name(context,
                                                                quota_class)
        else:
            class_quotas = {}

        default_quotas = self.get_defaults(context, resources)

        for resource in resources.values():
            limit = quotas.get(resource.name, class_quotas.get(
                        resource.name, default_quotas[resource.name]))
            modified_quotas[resource.name] = dict(limit=limit)

            # Include usages if desired.  This is optional because one
            # internal consumer of this interface wants to access the
            # usages directly from inside a transaction.
            if usages:
                usage = usages.get(resource.name, {})
                modified_quotas[resource.name].update(
                    in_use=usage.get('in_use', 0),
                    )

            # Initialize remains quotas with the default limits.
            if remains:
                modified_quotas[resource.name].update(remains=limit)

        if remains:
            # Get all user quotas for a project and subtract their limits
            # from the class limits to get the remains. For example, if the
            # class/default is 20 and there are two users each with quota of 5,
            # then there is quota of 10 left to give out.
            all_quotas = objects.Quotas.get_all(context, project_id)
            for quota in all_quotas:
                if quota.resource in modified_quotas:
                    modified_quotas[quota.resource]['remains'] -= \
                            quota.hard_limit

        return modified_quotas

    def _get_usages(self, context, resources, project_id, user_id=None):
        """Get usages of specified resources.

        This function is called to get resource usages for validating quota
        limit creates or updates in the os-quota-sets API and for displaying
        resource usages in the os-used-limits API. This function is not used
        for checking resource usage against quota limits.

        :param context: The request context for access checks
        :param resources: The dict of Resources for which to get usages
        :param project_id: The project_id for scoping the usage count
        :param user_id: Optional user_id for scoping the usage count
        :returns: A dict containing resources and their usage information,
                  for example:
                  {'project_id': 'project-uuid',
                   'user_id': 'user-uuid',
                   'instances': {'in_use': 5}}
        """
        usages = {}
        for resource in resources.values():
            # NOTE(melwitt): We should skip resources that are not countable,
            # such as AbsoluteResources.
            if not isinstance(resource, CountableResource):
                continue
            if resource.name in usages:
                # This is needed because for any of the resources:
                # ('instances', 'cores', 'ram'), they are counted at the same
                # time for efficiency (query the instances table once instead
                # of multiple times). So, a count of any one of them contains
                # counts for the others and we can avoid re-counting things.
                continue
            if resource.name in ('key_pairs', 'server_group_members'):
                # These per user resources are special cases whose usages
                # are not considered when validating limit create/update or
                # displaying used limits. They are always zero.
                usages[resource.name] = {'in_use': 0}
            else:
                if (
                    resource.name in
                    main_db_api.quota_get_per_project_resources()
                ):
                    count = resource.count_as_dict(context, project_id)
                    key = 'project'
                else:
                    # NOTE(melwitt): This assumes a specific signature for
                    # count_as_dict(). Usages used to be records in the
                    # database but now we are counting resources. The
                    # count_as_dict() function signature needs to match this
                    # call, else it should get a conditional in this function.
                    count = resource.count_as_dict(context, project_id,
                                                   user_id=user_id)
                    key = 'user' if user_id else 'project'
                # Example count_as_dict() return value:
                #   {'project': {'instances': 5},
                #    'user': {'instances': 2}}
                counted_resources = count[key].keys()
                for res in counted_resources:
                    count_value = count[key][res]
                    usages[res] = {'in_use': count_value}
        return usages

    def get_user_quotas(self, context, resources, project_id, user_id,
                        quota_class=None,
                        usages=True, project_quotas=None,
                        user_quotas=None):
        """Given a list of resources, retrieve the quotas for the given
        user and project.

        :param context: The request context, for access checks.
        :param resources: A dictionary of the registered resources.
        :param project_id: The ID of the project to return quotas for.
        :param user_id: The ID of the user to return quotas for.
        :param quota_class: If project_id != context.project_id, the
                            quota class cannot be determined.  This
                            parameter allows it to be specified.  It
                            will be ignored if project_id ==
                            context.project_id.
        :param usages: If True, the current counts will also be returned.
        :param project_quotas: Quotas dictionary for the specified project.
        :param user_quotas: Quotas dictionary for the specified project
                            and user.
        """
        if user_quotas:
            user_quotas = user_quotas.copy()
        else:
            user_quotas = objects.Quotas.get_all_by_project_and_user(
                context, project_id, user_id)
        # Use the project quota for default user quota.
        proj_quotas = project_quotas or objects.Quotas.get_all_by_project(
            context, project_id)
        for key, value in proj_quotas.items():
            if key not in user_quotas.keys():
                user_quotas[key] = value
        user_usages = {}
        if usages:
            user_usages = self._get_usages(context, resources, project_id,
                                           user_id=user_id)
        return self._process_quotas(context, resources, project_id,
                                    user_quotas, quota_class,
                                    usages=user_usages)

    def get_project_quotas(self, context, resources, project_id,
                           quota_class=None,
                           usages=True, remains=False, project_quotas=None):
        """Given a list of resources, retrieve the quotas for the given
        project.

        :param context: The request context, for access checks.
        :param resources: A dictionary of the registered resources.
        :param project_id: The ID of the project to return quotas for.
        :param quota_class: If project_id != context.project_id, the
                            quota class cannot be determined.  This
                            parameter allows it to be specified.  It
                            will be ignored if project_id ==
                            context.project_id.
        :param usages: If True, the current counts will also be returned.
        :param remains: If True, the current remains of the project will
                        will be returned.
        :param project_quotas: Quotas dictionary for the specified project.
        """
        project_quotas = project_quotas or objects.Quotas.get_all_by_project(
            context, project_id)
        project_usages = {}
        if usages:
            project_usages = self._get_usages(context, resources, project_id)
        return self._process_quotas(context, resources, project_id,
                                    project_quotas, quota_class,
                                    usages=project_usages,
                                    remains=remains)

    def _is_unlimited_value(self, v):
        """A helper method to check for unlimited value.
        """

        return v <= self.UNLIMITED_VALUE

    def _sum_quota_values(self, v1, v2):
        """A helper method that handles unlimited values when performing
        sum operation.
        """

        if self._is_unlimited_value(v1) or self._is_unlimited_value(v2):
            return self.UNLIMITED_VALUE
        return v1 + v2

    def _sub_quota_values(self, v1, v2):
        """A helper method that handles unlimited values when performing
        subtraction operation.
        """

        if self._is_unlimited_value(v1) or self._is_unlimited_value(v2):
            return self.UNLIMITED_VALUE
        return v1 - v2

    def get_settable_quotas(self, context, resources, project_id,
                            user_id=None):
        """Given a list of resources, retrieve the range of settable quotas for
        the given user or project.

        :param context: The request context, for access checks.
        :param resources: A dictionary of the registered resources.
        :param project_id: The ID of the project to return quotas for.
        :param user_id: The ID of the user to return quotas for.
        """

        settable_quotas = {}
        db_proj_quotas = objects.Quotas.get_all_by_project(context, project_id)
        project_quotas = self.get_project_quotas(context, resources,
                                                 project_id, remains=True,
                                                 project_quotas=db_proj_quotas)
        if user_id:
            setted_quotas = objects.Quotas.get_all_by_project_and_user(
                context, project_id, user_id)
            user_quotas = self.get_user_quotas(context, resources,
                                               project_id, user_id,
                                               project_quotas=db_proj_quotas,
                                               user_quotas=setted_quotas)
            for key, value in user_quotas.items():
                # Maximum is the remaining quota for a project (class/default
                # minus the sum of all user quotas in the project), plus the
                # given user's quota. So if the class/default is 20 and there
                # are two users each with quota of 5, then there is quota of
                # 10 remaining. The given user currently has quota of 5, so
                # the maximum you could update their quota to would be 15.
                # Class/default 20 - currently used in project 10 + current
                # user 5 = 15.
                maximum = \
                    self._sum_quota_values(project_quotas[key]['remains'],
                                           setted_quotas.get(key, 0))
                # This function is called for the quota_sets api and the
                # corresponding nova-manage command. The idea is when someone
                # attempts to update a quota, the value chosen must be at least
                # as much as the current usage and less than or equal to the
                # project limit less the sum of existing per user limits.
                minimum = value['in_use']
                settable_quotas[key] = {'minimum': minimum, 'maximum': maximum}
        else:
            for key, value in project_quotas.items():
                minimum = \
                    max(int(self._sub_quota_values(value['limit'],
                                                   value['remains'])),
                        int(value['in_use']))
                settable_quotas[key] = {'minimum': minimum, 'maximum': -1}
        return settable_quotas

    def _get_quotas(self, context, resources, keys, project_id=None,
                    user_id=None, project_quotas=None):
        """A helper method which retrieves the quotas for the specific
        resources identified by keys, and which apply to the current
        context.

        :param context: The request context, for access checks.
        :param resources: A dictionary of the registered resources.
        :param keys: A list of the desired quotas to retrieve.
        :param project_id: Specify the project_id if current context
                           is admin and admin wants to impact on
                           common user's tenant.
        :param user_id: Specify the user_id if current context
                        is admin and admin wants to impact on
                        common user.
        :param project_quotas: Quotas dictionary for the specified project.
        """

        # Filter resources
        desired = set(keys)
        sub_resources = {k: v for k, v in resources.items() if k in desired}

        # Make sure we accounted for all of them...
        if len(keys) != len(sub_resources):
            unknown = desired - set(sub_resources.keys())
            raise exception.QuotaResourceUnknown(unknown=sorted(unknown))

        if user_id:
            LOG.debug('Getting quotas for user %(user_id)s and project '
                      '%(project_id)s. Resources: %(keys)s',
                      {'user_id': user_id, 'project_id': project_id,
                       'keys': keys})
            # Grab and return the quotas (without usages)
            quotas = self.get_user_quotas(context, sub_resources,
                                          project_id, user_id,
                                          context.quota_class, usages=False,
                                          project_quotas=project_quotas)
        else:
            LOG.debug('Getting quotas for project %(project_id)s. Resources: '
                      '%(keys)s', {'project_id': project_id, 'keys': keys})
            # Grab and return the quotas (without usages)
            quotas = self.get_project_quotas(context, sub_resources,
                                             project_id,
                                             context.quota_class,
                                             usages=False,
                                             project_quotas=project_quotas)

        return {k: v['limit'] for k, v in quotas.items()}

    def limit_check(self, context, resources, values, project_id=None,
                    user_id=None):
        """Check simple quota limits.

        For limits--those quotas for which there is no usage
        synchronization function--this method checks that a set of
        proposed values are permitted by the limit restriction.

        This method will raise a QuotaResourceUnknown exception if a
        given resource is unknown or if it is not a simple limit
        resource.

        If any of the proposed values is over the defined quota, an
        OverQuota exception will be raised with the sorted list of the
        resources which are too high.  Otherwise, the method returns
        nothing.

        :param context: The request context, for access checks.
        :param resources: A dictionary of the registered resources.
        :param values: A dictionary of the values to check against the
                       quota.
        :param project_id: Specify the project_id if current context
                           is admin and admin wants to impact on
                           common user's tenant.
        :param user_id: Specify the user_id if current context
                        is admin and admin wants to impact on
                        common user.
        """
        _valid_method_call_check_resources(values, 'check', resources)

        # Ensure no value is less than zero
        unders = [key for key, val in values.items() if val < 0]
        if unders:
            raise exception.InvalidQuotaValue(unders=sorted(unders))

        # If project_id is None, then we use the project_id in context
        if project_id is None:
            project_id = context.project_id
        # If user id is None, then we use the user_id in context
        if user_id is None:
            user_id = context.user_id

        # Get the applicable quotas
        project_quotas = objects.Quotas.get_all_by_project(context, project_id)
        quotas = self._get_quotas(context, resources, values.keys(),
                                  project_id=project_id,
                                  project_quotas=project_quotas)
        user_quotas = self._get_quotas(context, resources, values.keys(),
                                       project_id=project_id,
                                       user_id=user_id,
                                       project_quotas=project_quotas)

        # Check the quotas and construct a list of the resources that
        # would be put over limit by the desired values
        overs = [key for key, val in values.items()
                 if quotas[key] >= 0 and quotas[key] < val or
                 (user_quotas[key] >= 0 and user_quotas[key] < val)]
        if overs:
            headroom = {}
            for key in overs:
                headroom[key] = min(
                    val for val in (quotas.get(key), project_quotas.get(key))
                    if val is not None
                )
            raise exception.OverQuota(overs=sorted(overs), quotas=quotas,
                                      usages={}, headroom=headroom)

    def limit_check_project_and_user(self, context, resources,
                                     project_values=None, user_values=None,
                                     project_id=None, user_id=None):
        """Check values (usage + desired delta) against quota limits.

        For limits--this method checks that a set of
        proposed values are permitted by the limit restriction.

        This method will raise a QuotaResourceUnknown exception if a
        given resource is unknown or if it is not a simple limit
        resource.

        If any of the proposed values is over the defined quota, an
        OverQuota exception will be raised with the sorted list of the
        resources which are too high.  Otherwise, the method returns
        nothing.

        :param context: The request context, for access checks
        :param resources: A dictionary of the registered resources
        :param project_values: Optional dict containing the resource values to
                            check against project quota,
                            e.g. {'instances': 1, 'cores': 2, 'memory_mb': 512}
        :param user_values: Optional dict containing the resource values to
                            check against user quota,
                            e.g. {'instances': 1, 'cores': 2, 'memory_mb': 512}
        :param project_id: Optional project_id for scoping the limit check to a
                           different project than in the context
        :param user_id: Optional user_id for scoping the limit check to a
                        different user than in the context
        """
        if project_values is None:
            project_values = {}
        if user_values is None:
            user_values = {}

        _valid_method_call_check_resources(project_values, 'check', resources)
        _valid_method_call_check_resources(user_values, 'check', resources)

        if not any([project_values, user_values]):
            raise exception.Invalid(
                'Must specify at least one of project_values or user_values '
                'for the limit check.')

        # Ensure no value is less than zero
        for vals in (project_values, user_values):
            unders = [key for key, val in vals.items() if val < 0]
            if unders:
                raise exception.InvalidQuotaValue(unders=sorted(unders))

        # Get a set of all keys for calling _get_quotas() so we get all of the
        # resource limits we need.
        all_keys = set(project_values).union(user_values)

        # Keys that are in both project_values and user_values need to be
        # checked against project quota and user quota, respectively.
        # Keys that are not in both only need to be checked against project
        # quota or user quota, if it is defined. Separate the keys that don't
        # need to be checked against both quotas, merge them into one dict,
        # and remove them from project_values and user_values.
        keys_to_merge = set(project_values).symmetric_difference(user_values)
        merged_values = {}
        for key in keys_to_merge:
            # The key will be either in project_values or user_values based on
            # the earlier symmetric_difference. Default to 0 in case the found
            # value is 0 and won't take precedence over a None default.
            merged_values[key] = (project_values.get(key, 0) or
                                  user_values.get(key, 0))
            project_values.pop(key, None)
            user_values.pop(key, None)

        # If project_id is None, then we use the project_id in context
        if project_id is None:
            project_id = context.project_id
        # If user id is None, then we use the user_id in context
        if user_id is None:
            user_id = context.user_id

        # Get the applicable quotas. They will be merged together (taking the
        # min limit) if project_values and user_values were not specified
        # together.

        # per project quota limits (quotas that have no concept of
        # user-scoping: <none>)
        project_quotas = objects.Quotas.get_all_by_project(context, project_id)
        # per user quotas, project quota limits (for quotas that have
        # user-scoping, limits for the project)
        quotas = self._get_quotas(context, resources, all_keys,
                                  project_id=project_id,
                                  project_quotas=project_quotas)
        # per user quotas, user quota limits (for quotas that have
        # user-scoping, the limits for the user)
        user_quotas = self._get_quotas(context, resources, all_keys,
                                       project_id=project_id,
                                       user_id=user_id,
                                       project_quotas=project_quotas)

        if merged_values:
            # This is for resources that are not counted across a project and
            # must pass both the quota for the project and the quota for the
            # user.
            # Combine per user project quotas and user_quotas for use in the
            # checks, taking the minimum limit between the two.
            merged_quotas = copy.deepcopy(quotas)
            for k, v in user_quotas.items():
                if k in merged_quotas:
                    merged_quotas[k] = min(merged_quotas[k], v)
                else:
                    merged_quotas[k] = v

            # Check the quotas and construct a list of the resources that
            # would be put over limit by the desired values
            overs = [key for key, val in merged_values.items()
                     if merged_quotas[key] >= 0 and merged_quotas[key] < val]
            if overs:
                headroom = {}
                for key in overs:
                    headroom[key] = merged_quotas[key]
                raise exception.OverQuota(overs=sorted(overs),
                                          quotas=merged_quotas, usages={},
                                          headroom=headroom)

        # This is for resources that are counted across a project and
        # across a user (instances, cores, ram, server_groups). The
        # project_values must pass the quota for the project and the
        # user_values must pass the quota for the user.
        over_user_quota = False
        overs = []
        for key in user_values.keys():
            # project_values and user_values should contain the same keys or
            # be empty after the keys in the symmetric_difference were removed
            # from both dicts.
            if quotas[key] >= 0 and quotas[key] < project_values[key]:
                overs.append(key)
            elif (user_quotas[key] >= 0 and
                  user_quotas[key] < user_values[key]):
                overs.append(key)
                over_user_quota = True
        if overs:
            quotas_exceeded = user_quotas if over_user_quota else quotas
            headroom = {}
            for key in overs:
                headroom[key] = quotas_exceeded[key]
            raise exception.OverQuota(overs=sorted(overs),
                                      quotas=quotas_exceeded, usages={},
                                      headroom=headroom)


class NoopQuotaDriver(object):
    """Driver that turns quotas calls into no-ops and pretends that quotas
    for all resources are unlimited. This can be used if you do not
    wish to have any quota checking.
    """

    def get_reserved(self):
        # Noop has always returned -1 for reserved
        return -1

    def get_defaults(self, context, resources):
        """Given a list of resources, retrieve the default quotas.

        :param context: The request context, for access checks.
        :param resources: A dictionary of the registered resources.
        """
        quotas = {}
        for resource in resources.values():
            quotas[resource.name] = -1
        return quotas

    def get_class_quotas(self, context, resources, quota_class):
        """Given a list of resources, retrieve the quotas for the given
        quota class.

        :param context: The request context, for access checks.
        :param resources: A dictionary of the registered resources.
        :param quota_class: The name of the quota class to return
                            quotas for.
        """
        quotas = {}
        for resource in resources.values():
            quotas[resource.name] = -1
        return quotas

    def _get_noop_quotas(self, resources, usages=None, remains=False):
        quotas = {}
        for resource in resources.values():
            quotas[resource.name] = {}
            quotas[resource.name]['limit'] = -1
            if usages:
                quotas[resource.name]['in_use'] = -1
            if remains:
                quotas[resource.name]['remains'] = -1
        return quotas

    def get_user_quotas(self, context, resources, project_id, user_id,
                        quota_class=None,
                        usages=True):
        """Given a list of resources, retrieve the quotas for the given
        user and project.

        :param context: The request context, for access checks.
        :param resources: A dictionary of the registered resources.
        :param project_id: The ID of the project to return quotas for.
        :param user_id: The ID of the user to return quotas for.
        :param quota_class: If project_id != context.project_id, the
                            quota class cannot be determined.  This
                            parameter allows it to be specified.  It
                            will be ignored if project_id ==
                            context.project_id.
        :param usages: If True, the current counts will also be returned.
        """
        return self._get_noop_quotas(resources, usages=usages)

    def get_project_quotas(self, context, resources, project_id,
                           quota_class=None,
                           usages=True, remains=False):
        """Given a list of resources, retrieve the quotas for the given
        project.

        :param context: The request context, for access checks.
        :param resources: A dictionary of the registered resources.
        :param project_id: The ID of the project to return quotas for.
        :param quota_class: If project_id != context.project_id, the
                            quota class cannot be determined.  This
                            parameter allows it to be specified.  It
                            will be ignored if project_id ==
                            context.project_id.
        :param usages: If True, the current counts will also be returned.
        :param remains: If True, the current remains of the project will
                        will be returned.
        """
        return self._get_noop_quotas(resources, usages=usages, remains=remains)

    def get_settable_quotas(self, context, resources, project_id,
                            user_id=None):
        """Given a list of resources, retrieve the range of settable quotas for
        the given user or project.

        :param context: The request context, for access checks.
        :param resources: A dictionary of the registered resources.
        :param project_id: The ID of the project to return quotas for.
        :param user_id: The ID of the user to return quotas for.
        """
        quotas = {}
        for resource in resources.values():
            quotas[resource.name] = {'minimum': 0, 'maximum': -1}
        return quotas

    def limit_check(self, context, resources, values, project_id=None,
                    user_id=None):
        """Check simple quota limits.

        For limits--those quotas for which there is no usage
        synchronization function--this method checks that a set of
        proposed values are permitted by the limit restriction.

        This method will raise a QuotaResourceUnknown exception if a
        given resource is unknown or if it is not a simple limit
        resource.

        If any of the proposed values is over the defined quota, an
        OverQuota exception will be raised with the sorted list of the
        resources which are too high.  Otherwise, the method returns
        nothing.

        :param context: The request context, for access checks.
        :param resources: A dictionary of the registered resources.
        :param values: A dictionary of the values to check against the
                       quota.
        :param project_id: Specify the project_id if current context
                           is admin and admin wants to impact on
                           common user's tenant.
        :param user_id: Specify the user_id if current context
                        is admin and admin wants to impact on
                        common user.
        """
        pass

    def limit_check_project_and_user(self, context, resources,
                                     project_values=None, user_values=None,
                                     project_id=None, user_id=None):
        """Check values against quota limits.

        For limits--this method checks that a set of
        proposed values are permitted by the limit restriction.

        This method will raise a QuotaResourceUnknown exception if a
        given resource is unknown or if it is not a simple limit
        resource.

        If any of the proposed values is over the defined quota, an
        OverQuota exception will be raised with the sorted list of the
        resources which are too high.  Otherwise, the method returns
        nothing.

        :param context: The request context, for access checks
        :param resources: A dictionary of the registered resources
        :param project_values: Optional dict containing the resource values to
                            check against project quota,
                            e.g. {'instances': 1, 'cores': 2, 'memory_mb': 512}
        :param user_values: Optional dict containing the resource values to
                            check against user quota,
                            e.g. {'instances': 1, 'cores': 2, 'memory_mb': 512}
        :param project_id: Optional project_id for scoping the limit check to a
                           different project than in the context
        :param user_id: Optional user_id for scoping the limit check to a
                        different user than in the context
        """
        pass


class UnifiedLimitsDriver(NoopQuotaDriver):
    """Ease migration to new unified limits code.

    Help ease migration to unified limits by ensuring the old code
    paths still work with unified limits. Eventually the expectation is
    all this legacy quota code will go away, leaving the new simpler code
    """

    def __init__(self):
        LOG.warning("The Unified Limits Quota Driver is experimental and "
                    "is under active development. Do not use this driver.")

    def get_reserved(self):
        # To make unified limits APIs the same as the DB driver, return 0
        return 0

    def get_class_quotas(self, context, resources, quota_class):
        """Given a list of resources, retrieve the quotas for the given
        quota class.

        :param context: The request context, for access checks.
        :param resources: A dictionary of the registered resources.
        :param quota_class: Placeholder, we always assume default quota class.
        """
        # NOTE(johngarbutt): ignoring quota_class, as ignored in noop driver
        return self.get_defaults(context, resources)

    def get_defaults(self, context, resources):
        local_limits = local_limit.get_legacy_default_limits()
        # Note we get 0 if there is no registered limit,
        # to mirror oslo_limit behaviour when there is no registered limit
        placement_limits = placement_limit.get_legacy_default_limits()
        quotas = {}
        for resource in resources.values():
            if resource.name in placement_limits:
                quotas[resource.name] = placement_limits[resource.name]
            else:
                # return -1 for things like security_group_rules
                # that are neither a keystone limit or a local limit
                quotas[resource.name] = local_limits.get(resource.name, -1)

        return quotas

    def get_project_quotas(self, context, resources, project_id,
                           quota_class=None,
                           usages=True, remains=False):
        if quota_class is not None:
            raise NotImplementedError("quota_class")

        if remains:
            raise NotImplementedError("remains")

        local_limits = local_limit.get_legacy_default_limits()
        # keystone limits always returns core, ram and instances
        # if nothing set in keystone, we get back 0, i.e. don't allow
        placement_limits = placement_limit.get_legacy_project_limits(
            project_id)

        project_quotas = {}
        for resource in resources.values():
            if resource.name in placement_limits:
                limit = placement_limits[resource.name]
            else:
                # return -1 for things like security_group_rules
                # that are neither a keystone limit or a local limit
                limit = local_limits.get(resource.name, -1)
            project_quotas[resource.name] = {"limit": limit}

        if usages:
            local_in_use = local_limit.get_in_use(context, project_id)
            p_in_use = placement_limit.get_legacy_counts(context, project_id)

            for resource in resources.values():
                # default to 0 for resources that are deprecated,
                # i.e. not in keystone or local limits, such that we
                # are API compatible with what was returned with
                # the db driver, even though noop driver returned -1
                usage_count = 0
                if resource.name in local_in_use:
                    usage_count = local_in_use[resource.name]
                if resource.name in p_in_use:
                    usage_count = p_in_use[resource.name]
                project_quotas[resource.name]["in_use"] = usage_count

        return project_quotas

    def get_user_quotas(self, context, resources, project_id, user_id,
                        quota_class=None, usages=True):
        return self.get_project_quotas(context, resources, project_id,
                                       quota_class, usages)


class BaseResource(object):
    """Describe a single resource for quota checking."""

    def __init__(self, name, flag=None):
        """Initializes a Resource.

        :param name: The name of the resource, i.e., "instances".
        :param flag: The name of the flag or configuration option
                     which specifies the default value of the quota
                     for this resource.
        """

        self.name = name
        self.flag = flag

    @property
    def default(self):
        """Return the default value of the quota."""
        return CONF.quota[self.flag] if self.flag else -1


class AbsoluteResource(BaseResource):
    """Describe a resource that does not correspond to database objects."""
    valid_method = 'check'


class CountableResource(AbsoluteResource):
    """Describe a resource where the counts aren't based solely on the
    project ID.
    """

    def __init__(self, name, count_as_dict, flag=None):
        """Initializes a CountableResource.

        Countable resources are those resources which directly
        correspond to objects in the database, but for which a count
        by project ID is inappropriate e.g. keypairs
        A CountableResource must be constructed with a counting
        function, which will be called to determine the current counts
        of the resource.

        The counting function will be passed the context, along with
        the extra positional and keyword arguments that are passed to
        Quota.count_as_dict().  It should return a dict specifying the
        count scoped to a project and/or a user.

        Example count of instances, cores, or ram returned as a rollup
        of all the resources since we only want to query the instances
        table once, not multiple times, for each resource.
        Instances, cores, and ram are counted across a project and
        across a user:

            {'project': {'instances': 5, 'cores': 8, 'ram': 4096},
             'user': {'instances': 1, 'cores': 2, 'ram': 512}}

        Example count of server groups keeping a consistent format.
        Server groups are counted across a project and across a user:

            {'project': {'server_groups': 7},
             'user': {'server_groups': 2}}

        Example count of key pairs keeping a consistent format.
        Key pairs are counted across a user only:

            {'user': {'key_pairs': 5}}

        Note that this counting is not performed in a transaction-safe
        manner.  This resource class is a temporary measure to provide
        required functionality, until a better approach to solving
        this problem can be evolved.

        :param name: The name of the resource, i.e., "instances".
        :param count_as_dict: A callable which returns the count of the
                              resource as a dict.  The arguments passed are as
                              described above.
        :param flag: The name of the flag or configuration option
                     which specifies the default value of the quota
                     for this resource.
        """

        super(CountableResource, self).__init__(name, flag=flag)
        self.count_as_dict = count_as_dict


class QuotaEngine(object):
    """Represent the set of recognized quotas."""

    def __init__(self, quota_driver=None, resources=None):
        """Initialize a Quota object.

        :param quota_driver: a QuotaDriver object (only used in testing. if
                             None (default), instantiates a driver from the
                             CONF.quota.driver option)
        :param resources: iterable of Resource objects
        """
        resources = resources or []
        self._resources = {
            resource.name: resource for resource in resources
        }
        # NOTE(mriedem): quota_driver is ever only supplied in tests with a
        # fake driver.
        self.__driver_override = quota_driver
        self.__driver = None
        self.__driver_name = None

    @property
    def _driver(self):
        if self.__driver_override:
            return self.__driver_override

        # NOTE(johngarbutt) to allow unit tests to change the driver by
        # simply overriding config, double check if we have the correct
        # driver cached before we return the currently cached driver
        driver_name_in_config = CONF.quota.driver
        if self.__driver_name != driver_name_in_config:
            self.__driver = importutils.import_object(driver_name_in_config)
            self.__driver_name = driver_name_in_config

        return self.__driver

    def get_defaults(self, context):
        """Retrieve the default quotas.

        :param context: The request context, for access checks.
        """

        return self._driver.get_defaults(context, self._resources)

    def get_class_quotas(self, context, quota_class):
        """Retrieve the quotas for the given quota class.

        :param context: The request context, for access checks.
        :param quota_class: The name of the quota class to return
                            quotas for.
        """

        return self._driver.get_class_quotas(context, self._resources,
                                             quota_class)

    def get_user_quotas(self, context, project_id, user_id, quota_class=None,
                        usages=True):
        """Retrieve the quotas for the given user and project.

        :param context: The request context, for access checks.
        :param project_id: The ID of the project to return quotas for.
        :param user_id: The ID of the user to return quotas for.
        :param quota_class: If project_id != context.project_id, the
                            quota class cannot be determined.  This
                            parameter allows it to be specified.
        :param usages: If True, the current counts will also be returned.
        """

        return self._driver.get_user_quotas(context, self._resources,
                                            project_id, user_id,
                                            quota_class=quota_class,
                                            usages=usages)

    def get_project_quotas(self, context, project_id, quota_class=None,
                           usages=True, remains=False):
        """Retrieve the quotas for the given project.

        :param context: The request context, for access checks.
        :param project_id: The ID of the project to return quotas for.
        :param quota_class: If project_id != context.project_id, the
                            quota class cannot be determined.  This
                            parameter allows it to be specified.
        :param usages: If True, the current counts will also be returned.
        :param remains: If True, the current remains of the project will
                        will be returned.
        """

        return self._driver.get_project_quotas(context, self._resources,
                                              project_id,
                                              quota_class=quota_class,
                                              usages=usages,
                                              remains=remains)

    def get_settable_quotas(self, context, project_id, user_id=None):
        """Given a list of resources, retrieve the range of settable quotas for
        the given user or project.

        :param context: The request context, for access checks.
        :param project_id: The ID of the project to return quotas for.
        :param user_id: The ID of the user to return quotas for.
        """

        return self._driver.get_settable_quotas(context, self._resources,
                                                project_id,
                                                user_id=user_id)

    def count_as_dict(self, context, resource, *args, **kwargs):
        """Count a resource and return a dict.

        For countable resources, invokes the count_as_dict() function and
        returns its result.  Arguments following the context and
        resource are passed directly to the count function declared by
        the resource.

        :param context: The request context, for access checks.
        :param resource: The name of the resource, as a string.
        :returns: A dict containing the count(s) for the resource, for example:
                    {'project': {'instances': 2, 'cores': 4, 'ram': 1024},
                     'user': {'instances': 1, 'cores': 2, 'ram': 512}}

                  another example:
                    {'user': {'key_pairs': 5}}
        """

        # Get the resource
        res = self._resources.get(resource)
        if not res or not hasattr(res, 'count_as_dict'):
            raise exception.QuotaResourceUnknown(unknown=[resource])

        return res.count_as_dict(context, *args, **kwargs)

    # TODO(melwitt): This can be removed once no old code can call
    # limit_check(). It will be replaced with limit_check_project_and_user().
    def limit_check(self, context, project_id=None, user_id=None, **values):
        """Check simple quota limits.

        For limits--those quotas for which there is no usage
        synchronization function--this method checks that a set of
        proposed values are permitted by the limit restriction.  The
        values to check are given as keyword arguments, where the key
        identifies the specific quota limit to check, and the value is
        the proposed value.

        This method will raise a QuotaResourceUnknown exception if a
        given resource is unknown or if it is not a simple limit
        resource.

        If any of the proposed values is over the defined quota, an
        OverQuota exception will be raised with the sorted list of the
        resources which are too high.  Otherwise, the method returns
        nothing.

        :param context: The request context, for access checks.
        :param project_id: Specify the project_id if current context
                           is admin and admin wants to impact on
                           common user's tenant.
        :param user_id: Specify the user_id if current context
                        is admin and admin wants to impact on
                        common user.
        """

        return self._driver.limit_check(context, self._resources, values,
                                        project_id=project_id, user_id=user_id)

    def limit_check_project_and_user(self, context, project_values=None,
                                     user_values=None, project_id=None,
                                     user_id=None):
        """Check values against quota limits.

        For limits--this method checks that a set of
        proposed values are permitted by the limit restriction.

        This method will raise a QuotaResourceUnknown exception if a
        given resource is unknown or if it is not a simple limit
        resource.

        If any of the proposed values is over the defined quota, an
        OverQuota exception will be raised with the sorted list of the
        resources which are too high.  Otherwise, the method returns
        nothing.

        :param context: The request context, for access checks
        :param project_values: Optional dict containing the resource values to
                            check against project quota,
                            e.g. {'instances': 1, 'cores': 2, 'memory_mb': 512}
        :param user_values: Optional dict containing the resource values to
                            check against user quota,
                            e.g. {'instances': 1, 'cores': 2, 'memory_mb': 512}
        :param project_id: Optional project_id for scoping the limit check to a
                           different project than in the context
        :param user_id: Optional user_id for scoping the limit check to a
                        different user than in the context
        """
        return self._driver.limit_check_project_and_user(
            context, self._resources, project_values=project_values,
            user_values=user_values, project_id=project_id, user_id=user_id)

    @property
    def resources(self):
        return sorted(self._resources.keys())

    def get_reserved(self):
        return self._driver.get_reserved()


@api_db_api.context_manager.reader
def _user_id_queued_for_delete_populated(context, project_id=None):
    """Determine whether user_id and queued_for_delete are set.

    This will be used to determine whether we need to fall back on
    the legacy quota counting method (if we cannot rely on counting
    instance mappings for the instance count). If any records with user_id=None
    and queued_for_delete=False are found, we need to fall back to the legacy
    counting method. If any records with queued_for_delete=None are found, we
    need to fall back to the legacy counting method.

    Note that this check specifies queued_for_deleted=False, which excludes
    deleted and SOFT_DELETED instances. The 'populate_user_id' data migration
    migrates SOFT_DELETED instances because they could be restored at any time
    in the future. However, for this quota-check-time method, it is acceptable
    to ignore SOFT_DELETED instances, since we just want to know if it is safe
    to use instance mappings to count instances at this point in time (and
    SOFT_DELETED instances do not count against quota limits).

    We also want to fall back to the legacy counting method if we detect any
    records that have not yet populated the queued_for_delete field. We do this
    instead of counting queued_for_delete=None records since that might not
    accurately reflect the project or project user's quota usage.

    :param project_id: The project to check
    :returns: True if user_id is set for all non-deleted instances and
              queued_for_delete is set for all instances, else False
    """
    user_id_not_populated = sql.and_(
        api_models.InstanceMapping.user_id == sql.null(),
        api_models.InstanceMapping.queued_for_delete == sql.false())
    # If either queued_for_delete or user_id are unmigrated, we will return
    # False.
    unmigrated_filter = sql.or_(
        api_models.InstanceMapping.queued_for_delete == sql.null(),
        user_id_not_populated)
    query = context.session.query(api_models.InstanceMapping).filter(
        unmigrated_filter)
    if project_id:
        query = query.filter_by(project_id=project_id)
    return not context.session.query(query.exists()).scalar()


def _keypair_get_count_by_user(context, user_id):
    count = objects.KeyPairList.get_count_by_user(context, user_id)
    return {'user': {'key_pairs': count}}


def _server_group_count_members_by_user_legacy(context, group, user_id):
    # NOTE(melwitt): This is mostly duplicated from
    # InstanceGroup.count_members_by_user() to query across multiple cells.
    # We need to be able to pass the correct cell context to
    # InstanceList.get_by_filters().
    # NOTE(melwitt): Counting across cells for instances means we will miss
    # counting resources if a cell is down.
    cell_mappings = objects.CellMappingList.get_all(context)
    greenthreads = []
    filters = {'deleted': False, 'user_id': user_id, 'uuid': group.members}
    for cell_mapping in cell_mappings:
        with nova_context.target_cell(context, cell_mapping) as cctxt:
            greenthreads.append(utils.spawn(
                objects.InstanceList.get_by_filters, cctxt, filters,
                expected_attrs=[]))
    instances = objects.InstanceList(objects=[])
    for greenthread in greenthreads:
        found = greenthread.wait()
        instances = instances + found
    # Count build requests using the same filters to catch group members
    # that are not yet creatd in a cell.
    # NOTE(mriedem): BuildRequestList.get_by_filters is not very efficient for
    # what we need and we can optimize this with a new query method.
    build_requests = objects.BuildRequestList.get_by_filters(context, filters)
    # Ignore any duplicates since build requests and instances can co-exist
    # for a short window of time after the instance is created in a cell but
    # before the build request is deleted.
    instance_uuids = [inst.uuid for inst in instances]
    count = len(instances)
    for build_request in build_requests:
        if build_request.instance_uuid not in instance_uuids:
            count += 1
    return {'user': {'server_group_members': count}}


def is_qfd_populated(context):
    """Check if user_id and queued_for_delete fields are populated.

    This method is related to counting quota usage from placement. It is not
    yet possible to count instances from placement, so in the meantime we can
    use instance mappings for counting. This method is used to determine
    whether the user_id and queued_for_delete columns are populated in the API
    database's instance_mappings table. Instance mapping records are not
    deleted from the database until the database is archived, so
    queued_for_delete tells us whether or not we should count them for instance
    quota usage. The user_id field enables us to scope instance quota usage to
    a user (legacy quota).

    Scoping instance quota to a user is only possible
    when counting quota usage from placement is configured and unified limits
    is not configured. When unified limits is configured, quotas are scoped
    only to projects.

    In the future when it is possible to count instance usage from placement,
    this method will no longer be needed.
    """
    global UID_QFD_POPULATED_CACHE_ALL
    if not UID_QFD_POPULATED_CACHE_ALL:
        LOG.debug('Checking whether user_id and queued_for_delete are '
                  'populated for all projects')
        UID_QFD_POPULATED_CACHE_ALL = _user_id_queued_for_delete_populated(
            context)

    return UID_QFD_POPULATED_CACHE_ALL


def _server_group_count_members_by_user(context, group, user_id):
    """Get the count of server group members for a group by user.

    :param context: The request context for database access
    :param group: The InstanceGroup object with members to count
    :param user_id: The user_id to count across
    :returns: A dict containing the user-scoped count. For example:

                {'user': 'server_group_members': <count across user>}}
    """
    # Because server group members quota counting is not scoped to a project,
    # but scoped to a particular InstanceGroup and user, we have no reasonable
    # way of pruning down our migration check to only a subset of all instance
    # mapping records.
    # So, we check whether user_id/queued_for_delete is populated for all
    # records and cache the result to prevent unnecessary checking once the
    # data migration has been completed.
    if is_qfd_populated(context):
        count = objects.InstanceMappingList.get_count_by_uuids_and_user(
            context, group.members, user_id)
        return {'user': {'server_group_members': count}}

    LOG.warning('Falling back to legacy quota counting method for server '
                'group members')
    return _server_group_count_members_by_user_legacy(context, group,
                                                      user_id)


def _instances_cores_ram_count_legacy(context, project_id, user_id=None):
    """Get the counts of instances, cores, and ram in cell databases.

    :param context: The request context for database access
    :param project_id: The project_id to count across
    :param user_id: The user_id to count across
    :returns: A dict containing the project-scoped counts and user-scoped
              counts if user_id is specified. For example:

                {'project': {'instances': <count across project>,
                             'cores': <count across project>,
                             'ram': <count across project>},
                 'user': {'instances': <count across user>,
                          'cores': <count across user>,
                          'ram': <count across user>}}
    """
    # NOTE(melwitt): Counting across cells for instances, cores, and ram means
    # we will miss counting resources if a cell is down.
    # NOTE(tssurya): We only go into those cells in which the tenant has
    # instances. We could optimize this to avoid the CellMappingList query
    # for single-cell deployments by checking the cell cache and only doing
    # this filtering if there is more than one non-cell0 cell.
    # TODO(tssurya): Consider adding a scatter_gather_cells_for_project
    # variant that makes this native to nova.context.
    if CONF.api.instance_list_per_project_cells:
        cell_mappings = objects.CellMappingList.get_by_project_id(
            context, project_id)
    else:
        nova_context.load_cells()
        cell_mappings = nova_context.CELLS
    results = nova_context.scatter_gather_cells(
        context, cell_mappings, nova_context.CELL_TIMEOUT,
        objects.InstanceList.get_counts, project_id, user_id=user_id)
    total_counts = {'project': {'instances': 0, 'cores': 0, 'ram': 0}}
    if user_id:
        total_counts['user'] = {'instances': 0, 'cores': 0, 'ram': 0}
    for result in results.values():
        if not nova_context.is_cell_failure_sentinel(result):
            for resource, count in result['project'].items():
                total_counts['project'][resource] += count
            if user_id:
                for resource, count in result['user'].items():
                    total_counts['user'][resource] += count
    return total_counts


def _cores_ram_count_placement(context, project_id, user_id=None):
    return report.report_client_singleton().get_usages_counts_for_quota(
        context, project_id, user_id=user_id)


def _instances_cores_ram_count_api_db_placement(context, project_id,
                                                user_id=None):
    # Will return a dict with format: {'project': {'instances': M},
    #                                  'user': {'instances': N}}
    # where the 'user' key is optional.
    total_counts = objects.InstanceMappingList.get_counts(context,
                                                          project_id,
                                                          user_id=user_id)
    cores_ram_counts = _cores_ram_count_placement(context, project_id,
                                                  user_id=user_id)
    total_counts['project'].update(cores_ram_counts['project'])
    if 'user' in total_counts:
        total_counts['user'].update(cores_ram_counts['user'])
    return total_counts


def _instances_cores_ram_count(context, project_id, user_id=None):
    """Get the counts of instances, cores, and ram.

    :param context: The request context for database access
    :param project_id: The project_id to count across
    :param user_id: The user_id to count across
    :returns: A dict containing the project-scoped counts and user-scoped
              counts if user_id is specified. For example:

                {'project': {'instances': <count across project>,
                             'cores': <count across project>,
                             'ram': <count across project>},
                 'user': {'instances': <count across user>,
                          'cores': <count across user>,
                          'ram': <count across user>}}
    """
    global UID_QFD_POPULATED_CACHE_BY_PROJECT
    if CONF.quota.count_usage_from_placement:
        # If a project has all user_id and queued_for_delete data populated,
        # cache the result to avoid needless database checking in the future.
        if (not UID_QFD_POPULATED_CACHE_ALL and
                project_id not in UID_QFD_POPULATED_CACHE_BY_PROJECT):
            LOG.debug('Checking whether user_id and queued_for_delete are '
                      'populated for project_id %s', project_id)
            uid_qfd_populated = _user_id_queued_for_delete_populated(
                context, project_id)
            if uid_qfd_populated:
                UID_QFD_POPULATED_CACHE_BY_PROJECT.add(project_id)
        else:
            uid_qfd_populated = True
        if uid_qfd_populated:
            return _instances_cores_ram_count_api_db_placement(context,
                                                               project_id,
                                                               user_id=user_id)
        LOG.warning('Falling back to legacy quota counting method for '
                    'instances, cores, and ram')
    return _instances_cores_ram_count_legacy(context, project_id,
                                             user_id=user_id)


def _server_group_count(context, project_id, user_id=None):
    """Get the counts of server groups in the database.

    :param context: The request context for database access
    :param project_id: The project_id to count across
    :param user_id: The user_id to count across
    :returns: A dict containing the project-scoped counts and user-scoped
              counts if user_id is specified. For example:

                {'project': {'server_groups': <count across project>},
                 'user': {'server_groups': <count across user>}}
    """
    return objects.InstanceGroupList.get_counts(context, project_id,
                                                user_id=user_id)


QUOTAS = QuotaEngine(
    resources=[
        CountableResource(
            'instances', _instances_cores_ram_count, 'instances'),
        CountableResource(
            'cores', _instances_cores_ram_count, 'cores'),
        CountableResource(
            'ram', _instances_cores_ram_count, 'ram'),
        AbsoluteResource(
            'metadata_items', 'metadata_items'),
        AbsoluteResource(
            'injected_files', 'injected_files'),
        AbsoluteResource(
            'injected_file_content_bytes', 'injected_file_content_bytes'),
        AbsoluteResource(
            'injected_file_path_bytes', 'injected_file_path_length'),
        CountableResource(
            'key_pairs', _keypair_get_count_by_user, 'key_pairs'),
        CountableResource(
            'server_groups', _server_group_count, 'server_groups'),
        CountableResource(
            'server_group_members', _server_group_count_members_by_user,
            'server_group_members'),
        # Deprecated nova-network quotas, retained to avoid changing API
        # responses
        AbsoluteResource('fixed_ips'),
        AbsoluteResource('floating_ips'),
        AbsoluteResource('security_groups'),
        AbsoluteResource('security_group_rules'),
    ],
)


def _valid_method_call_check_resource(name, method, resources):
    if name not in resources:
        raise exception.InvalidQuotaMethodUsage(method=method, res=name)
    res = resources[name]

    if res.valid_method != method:
        raise exception.InvalidQuotaMethodUsage(method=method, res=name)


def _valid_method_call_check_resources(resource_values, method, resources):
    """A method to check whether the resource can use the quota method.

    :param resource_values: Dict containing the resource names and values
    :param method: The quota method to check
    :param resources: Dict containing Resource objects to validate against
    """

    for name in resource_values.keys():
        _valid_method_call_check_resource(name, method, resources)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import fixtures
import mock
from oslo_utils.fixture import uuidsentinel as uuids

from nova.api.openstack.compute import flavor_access
from nova.policies import base as base_policy
from nova.policies import flavor_access as fa_policy
from nova.tests.unit.api.openstack import fakes
from nova.tests.unit import fake_flavor
from nova.tests.unit.policies import base


class FlavorAccessPolicyTest(base.BasePolicyTest):
    """Test Flavor Access APIs policies with all possible context.

    This class defines the set of context with different roles
    which are allowed and not allowed to pass the policy checks.
    With those set of context, it will call the API operation and
    verify the expected behaviour.
    """

    def setUp(self):
        super(FlavorAccessPolicyTest, self).setUp()
        self.controller = flavor_access.FlavorActionController()
        self.controller_index = flavor_access.FlavorAccessController()
        self.req = fakes.HTTPRequest.blank('')
        self.mock_get = self.useFixture(
            fixtures.MockPatch('nova.api.openstack.common.get_flavor')).mock
        uuid = uuids.fake_id
        self.flavor = fake_flavor.fake_flavor_obj(
                self.project_member_context,
                id=1, uuid=uuid, project_id=self.project_id,
                is_public=False)
        self.mock_get.return_value = self.flavor
        self.stub_out('nova.api.openstack.identity.verify_project_id',
                      lambda ctx, project_id: True)
        self.stub_out('nova.objects.flavor._get_projects_from_db',
                lambda context, flavorid: [])

        # With legacy rule and no scope checks, all admin is able to
        # add/remove flavor access to a tenant.
        self.admin_authorized_contexts = [
            self.legacy_admin_context, self.system_admin_context,
            self.project_admin_context]

        # With legacy rule, anyone can access flavor access info.
        self.admin_index_authorized_contexts = self.all_contexts

    def test_list_flavor_access_policy(self):
        rule_name = fa_policy.BASE_POLICY_NAME
        self.common_policy_auth(self.admin_index_authorized_contexts,
                                rule_name, self.controller_index.index,
                                self.req, '1')

    @mock.patch('nova.objects.Flavor.add_access')
    def test_add_tenant_access_policy(self, mock_add):
        rule_name = fa_policy.POLICY_ROOT % "add_tenant_access"
        self.common_policy_auth(self.admin_authorized_contexts,
                                rule_name,
                                self.controller._add_tenant_access,
                                self.req, '1',
                                body={'addTenantAccess': {'tenant': 't1'}})

    @mock.patch('nova.objects.Flavor.remove_access')
    def test_remove_tenant_access_policy(self, mock_remove):
        rule_name = fa_policy.POLICY_ROOT % "remove_tenant_access"
        self.common_policy_auth(self.admin_authorized_contexts,
                                rule_name,
                                self.controller._remove_tenant_access,
                                self.req, '1',
                                body={'removeTenantAccess': {'tenant': 't1'}})


class FlavorAccessNoLegacyNoScopeTest(FlavorAccessPolicyTest):
    """Test Flavor Access API policies with deprecated rules
    disabled, but scope checking still disabled.
    """

    without_deprecated_rules = True
    rules_without_deprecation = {
        fa_policy.POLICY_ROOT % "add_tenant_access":
            base_policy.ADMIN,
        fa_policy.POLICY_ROOT % "remove_tenant_access":
            base_policy.ADMIN,
        fa_policy.BASE_POLICY_NAME:
            base_policy.ADMIN}

    def setUp(self):
        super(FlavorAccessNoLegacyNoScopeTest, self).setUp()

        # with no legacy rule means all admin is able to list access info.
        self.admin_index_authorized_contexts = [
            self.legacy_admin_context, self.system_admin_context,
            self.project_admin_context]


class FlavorAccessScopeTypePolicyTest(FlavorAccessPolicyTest):
    """Test Flavor Access APIs policies with system scope enabled.

    This class set the nova.conf [oslo_policy] enforce_scope to True
    so that we can switch on the scope checking on oslo policy side.
    It defines the set of context with scoped token
    which are allowed and not allowed to pass the policy checks.
    With those set of context, it will run the API operation and
    verify the expected behaviour.
    """

    def setUp(self):
        super(FlavorAccessScopeTypePolicyTest, self).setUp()
        self.flags(enforce_scope=True, group="oslo_policy")

        # Scope checks remove project users power.
        self.admin_authorized_contexts = [
            self.system_admin_context]
        self.admin_index_authorized_contexts = [
            self.system_admin_context, self.system_member_context,
            self.system_reader_context, self.system_foo_context]


class FlavorAccessScopeTypeNoLegacyPolicyTest(FlavorAccessScopeTypePolicyTest):
    """Test FlavorAccess APIs policies with system scope enabled,
    and no more deprecated rules.
    """
    without_deprecated_rules = True
    rules_without_deprecation = {
        fa_policy.POLICY_ROOT % "add_tenant_access":
            base_policy.ADMIN,
        fa_policy.POLICY_ROOT % "remove_tenant_access":
            base_policy.ADMIN,
        fa_policy.BASE_POLICY_NAME:
            base_policy.ADMIN}

    def setUp(self):
        super(FlavorAccessScopeTypeNoLegacyPolicyTest, self).setUp()
        self.flags(enforce_scope=True, group="oslo_policy")
        self.admin_index_authorized_contexts = [
            self.system_admin_context]



#    Copyright 2013 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
from oslo_utils import uuidutils
from oslo_utils import versionutils

from nova import availability_zones
from nova import context as nova_context
from nova.db.main import api as db
from nova import exception
from nova.notifications.objects import base as notification
from nova.notifications.objects import service as service_notification
from nova import objects
from nova.objects import base
from nova.objects import fields


LOG = logging.getLogger(__name__)


# NOTE(danms): This is the global service version counter
SERVICE_VERSION = 61


# NOTE(danms): This is our SERVICE_VERSION history. The idea is that any
# time we bump the version, we will put an entry here to record the change,
# along with any pertinent data. For things that we can programatically
# detect that need a bump, we put something in _collect_things() below to
# assemble a dict of things we can check. For example, we pretty much always
# want to consider the compute RPC API version a thing that requires a service
# bump so that we can drive version pins from it. We could include other
# service RPC versions at some point, minimum object versions, etc.
#
# The TestServiceVersion test will fail if the calculated set of
# things differs from the value in the last item of the list below,
# indicating that a version bump is needed.
#
# Also note that there are other reasons we may want to bump this,
# which will not be caught by the test. An example of this would be
# triggering (or disabling) an online data migration once all services
# in the cluster are at the same level.
#
# If a version bump is required for something mechanical, just document
# that generic thing here (like compute RPC version bumps). No need to
# replicate the details from compute/rpcapi.py here. However, for more
# complex service interactions, extra detail should be provided
SERVICE_VERSION_HISTORY = (
    # Version 0: Pre-history
    {'compute_rpc': '4.0'},

    # Version 1: Introduction of SERVICE_VERSION
    {'compute_rpc': '4.4'},
    # Version 2: Compute RPC version 4.5
    {'compute_rpc': '4.5'},
    # Version 3: Compute RPC version 4.6
    {'compute_rpc': '4.6'},
    # Version 4: Add PciDevice.parent_addr (data migration needed)
    {'compute_rpc': '4.6'},
    # Version 5: Compute RPC version 4.7
    {'compute_rpc': '4.7'},
    # Version 6: Compute RPC version 4.8
    {'compute_rpc': '4.8'},
    # Version 7: Compute RPC version 4.9
    {'compute_rpc': '4.9'},
    # Version 8: Compute RPC version 4.10
    {'compute_rpc': '4.10'},
    # Version 9: Compute RPC version 4.11
    {'compute_rpc': '4.11'},
    # Version 10: Compute node conversion to Inventories
    {'compute_rpc': '4.11'},
    # Version 11: Compute RPC version 4.12
    {'compute_rpc': '4.12'},
    # Version 12: The network APIs and compute manager support a NetworkRequest
    # object where the network_id value is 'auto' or 'none'. BuildRequest
    # objects are populated by nova-api during instance boot.
    {'compute_rpc': '4.12'},
    # Version 13: Compute RPC version 4.13
    {'compute_rpc': '4.13'},
    # Version 14: The compute manager supports setting device tags.
    {'compute_rpc': '4.13'},
    # Version 15: Indicate that nova-conductor will stop a boot if BuildRequest
    # is deleted before RPC to nova-compute.
    {'compute_rpc': '4.13'},
    # Version 16: Indicate that nova-compute will refuse to start if it doesn't
    # have a placement section configured.
    {'compute_rpc': '4.13'},
    # Version 17: Add 'reserve_volume' to the boot from volume flow and
    # remove 'check_attach'. The service version bump is needed to fall back to
    # the old check in the API as the old computes fail if the volume is moved
    # to 'attaching' state by reserve.
    {'compute_rpc': '4.13'},
    # Version 18: Compute RPC version 4.14
    {'compute_rpc': '4.14'},
    # Version 19: Compute RPC version 4.15
    {'compute_rpc': '4.15'},
    # Version 20: Compute RPC version 4.16
    {'compute_rpc': '4.16'},
    # Version 21: Compute RPC version 4.17
    {'compute_rpc': '4.17'},
    # Version 22: A marker for the behaviour change of auto-healing code on the
    # compute host regarding allocations against an instance
    {'compute_rpc': '4.17'},
    # Version 23: Compute hosts allow pre-creation of the migration object
    # for cold migration.
    {'compute_rpc': '4.18'},
    # Version 24: Add support for Cinder v3 attach/detach API.
    {'compute_rpc': '4.18'},
    # Version 25: Compute hosts allow migration-based allocations
    # for live migration.
    {'compute_rpc': '4.18'},
    # Version 26: Adds a 'host_list' parameter to build_and_run_instance()
    {'compute_rpc': '4.19'},
    # Version 27: Compute RPC version 4.20; adds multiattach argument to
    # reserve_block_device_name().
    {'compute_rpc': '4.20'},
    # Version 28: Adds a 'host_list' parameter to prep_resize()
    {'compute_rpc': '4.21'},
    # Version 29: Compute RPC version 4.22
    {'compute_rpc': '4.22'},
    # Version 30: Compute RPC version 5.0
    {'compute_rpc': '5.0'},
    # Version 31: The compute manager checks if 'trusted_certs' are supported
    {'compute_rpc': '5.0'},
    # Version 32: Add 'file_backed_memory' support. The service version bump is
    # needed to allow the destination of a live migration to reject the
    # migration if 'file_backed_memory' is enabled and the source does not
    # support 'file_backed_memory'
    {'compute_rpc': '5.0'},
    # Version 33: Add support for check on the server group with
    # 'max_server_per_host' rules
    {'compute_rpc': '5.0'},
    # Version 34: Adds support to abort queued/preparing live migrations.
    {'compute_rpc': '5.0'},
    # Version 35: Indicates that nova-compute supports live migration with
    # ports bound early on the destination host using VIFMigrateData.
    {'compute_rpc': '5.0'},
    # Version 36: Indicates that nova-compute supports specifying volume
    # type when booting a volume-backed server.
    {'compute_rpc': '5.0'},
    # Version 37: prep_resize takes a RequestSpec object
    {'compute_rpc': '5.1'},
    # Version 38: set_host_enabled reflects COMPUTE_STATUS_DISABLED trait
    {'compute_rpc': '5.1'},
    # Version 39: resize_instance, finish_resize, revert_resize,
    # finish_revert_resize, unshelve_instance takes a RequestSpec object
    {'compute_rpc': '5.2'},
    # Version 40: Add migration and limits parameters to
    # check_can_live_migrate_destination(), new
    # drop_move_claim_at_destination() method, and numa_live_migration
    # parameter to check_can_live_migrate_source()
    {'compute_rpc': '5.3'},
    # Version 41: Add cache_images() to compute rpcapi (version 5.4)
    {'compute_rpc': '5.4'},
    # Version 42: Compute RPC version 5.5; +prep_snapshot_based_resize_at_dest
    {'compute_rpc': '5.5'},
    # Version 43: Compute RPC version 5.6: prep_snapshot_based_resize_at_source
    {'compute_rpc': '5.6'},
    # Version 44: Compute RPC version 5.7: finish_snapshot_based_resize_at_dest
    {'compute_rpc': '5.7'},
    # Version 45: Compute RPC v5.8: confirm_snapshot_based_resize_at_source
    {'compute_rpc': '5.8'},
    # Version 46: Compute RPC v5.9: revert_snapshot_based_resize_at_dest
    {'compute_rpc': '5.9'},
    # Version 47: Compute RPC v5.10:
    # finish_revert_snapshot_based_resize_at_source
    {'compute_rpc': '5.10'},
    # Version 48: Drivers report COMPUTE_SAME_HOST_COLD_MIGRATE trait.
    {'compute_rpc': '5.10'},
    # Version 49: Compute now support server move operations with qos ports
    {'compute_rpc': '5.10'},
    # Version 50: Compute RPC v5.11:
    # Add accel_uuids (accelerator requests) param to build_and_run_instance
    {'compute_rpc': '5.11'},
    # Version 51: Add support for live migration with vpmem
    {'compute_rpc': '5.11'},
    # Version 52: Add support for the 'mixed' CPU allocation policy
    {'compute_rpc': '5.11'},
    # Version 53: Compute RPC v5.12:
    # Add accel_uuids (accelerator requests) param to rebuild_instance
    {'compute_rpc': '5.12'},
    # Version 54: Compute RPC v5.13:
    # Add accel_uuids (accelerator requests) param to shelve_instance and
    # shelve_offload_instance and unshelve_instance
    {'compute_rpc': '5.13'},
    # Version 55: Compute RPC v5.13:
    # Add support for qos interface attach
    {'compute_rpc': '5.13'},
    # Version 56: Compute RPC v6.0:
    {'compute_rpc': '6.0'},
    # Version 57: Compute RPC v6.0:
    # Add support for vnic 'accelerator-direct'.
    {'compute_rpc': '6.0'},
    # Version 58: Compute RPC v6.0:
    # Add support for booting with neutron extended resource request
    {'compute_rpc': '6.0'},
    # Version 59: Compute RPC v6.0:
    # Add support for server move operations with neutron extended resource
    # request
    {'compute_rpc': '6.0'},
    # Version 60: Compute RPC v6.0:
    # Add support for interface attach operation with neutron extended resource
    # request
    {'compute_rpc': '6.0'},
    # Version 61: Compute RPC v6.0:
    # Add support for remotely-managed ports (vnic-type 'remote-managed')
    {'compute_rpc': '6.0'},
)

# This is used to raise an error at service startup if older than N-1 computes
# are detected. Update this at the beginning of every release cycle to point to
# the smallest service version that was added in N-1.
OLDEST_SUPPORTED_SERVICE_VERSION = 'Xena'
SERVICE_VERSION_ALIASES = {
    'Victoria': 52,
    'Wallaby': 54,
    'Xena': 57,
}


# TODO(berrange): Remove NovaObjectDictCompat
@base.NovaObjectRegistry.register
class Service(base.NovaPersistentObject, base.NovaObject,
              base.NovaObjectDictCompat):
    # Version 1.0: Initial version
    # Version 1.1: Added compute_node nested object
    # Version 1.2: String attributes updated to support unicode
    # Version 1.3: ComputeNode version 1.5
    # Version 1.4: Added use_slave to get_by_compute_host
    # Version 1.5: ComputeNode version 1.6
    # Version 1.6: ComputeNode version 1.7
    # Version 1.7: ComputeNode version 1.8
    # Version 1.8: ComputeNode version 1.9
    # Version 1.9: ComputeNode version 1.10
    # Version 1.10: Changes behaviour of loading compute_node
    # Version 1.11: Added get_by_host_and_binary
    # Version 1.12: ComputeNode version 1.11
    # Version 1.13: Added last_seen_up
    # Version 1.14: Added forced_down
    # Version 1.15: ComputeNode version 1.12
    # Version 1.16: Added version
    # Version 1.17: ComputeNode version 1.13
    # Version 1.18: ComputeNode version 1.14
    # Version 1.19: Added get_minimum_version()
    # Version 1.20: Added get_minimum_version_multi()
    # Version 1.21: Added uuid
    # Version 1.22: Added get_by_uuid()
    VERSION = '1.22'

    fields = {
        'id': fields.IntegerField(read_only=True),
        'uuid': fields.UUIDField(),
        'host': fields.StringField(nullable=True),
        'binary': fields.StringField(nullable=True),
        'topic': fields.StringField(nullable=True),
        'report_count': fields.IntegerField(),
        'disabled': fields.BooleanField(),
        'disabled_reason': fields.StringField(nullable=True),
        'availability_zone': fields.StringField(nullable=True),
        'compute_node': fields.ObjectField('ComputeNode'),
        'last_seen_up': fields.DateTimeField(nullable=True),
        'forced_down': fields.BooleanField(),
        'version': fields.IntegerField(),
    }

    _MIN_VERSION_CACHE = {}
    _SERVICE_VERSION_CACHING = False

    def __init__(self, *args, **kwargs):
        # NOTE(danms): We're going against the rules here and overriding
        # init. The reason is that we want to *ensure* that we're always
        # setting the current service version on our objects, overriding
        # whatever else might be set in the database, or otherwise (which
        # is the normal reason not to override init).
        #
        # We also need to do this here so that it's set on the client side
        # all the time, such that create() and save() operations will
        # include the current service version.
        if 'version' in kwargs:
            raise exception.ObjectActionError(
                action='init',
                reason='Version field is immutable')

        super(Service, self).__init__(*args, **kwargs)
        self.version = SERVICE_VERSION

    def obj_make_compatible_from_manifest(self, primitive, target_version,
                                          version_manifest):
        super(Service, self).obj_make_compatible_from_manifest(
            primitive, target_version, version_manifest)
        _target_version = versionutils.convert_version_to_tuple(target_version)
        if _target_version < (1, 21) and 'uuid' in primitive:
            del primitive['uuid']
        if _target_version < (1, 16) and 'version' in primitive:
            del primitive['version']
        if _target_version < (1, 14) and 'forced_down' in primitive:
            del primitive['forced_down']
        if _target_version < (1, 13) and 'last_seen_up' in primitive:
            del primitive['last_seen_up']
        if _target_version < (1, 10):
            # service.compute_node was not lazy-loaded, we need to provide it
            # when called
            self._do_compute_node(self._context, primitive,
                                  version_manifest)

    def _do_compute_node(self, context, primitive, version_manifest):
        try:
            target_version = version_manifest['ComputeNode']
            # NOTE(sbauza): Ironic deployments can have multiple
            # nodes for the same service, but for keeping same behaviour,
            # returning only the first elem of the list
            compute = objects.ComputeNodeList.get_all_by_host(
                context, primitive['host'])[0]
        except Exception:
            return
        primitive['compute_node'] = compute.obj_to_primitive(
            target_version=target_version,
            version_manifest=version_manifest)

    @staticmethod
    def _from_db_object(context, service, db_service):
        allow_missing = ('availability_zone',)
        for key in service.fields:
            if key in allow_missing and key not in db_service:
                continue
            if key == 'compute_node':
                #  NOTE(sbauza); We want to only lazy-load compute_node
                continue
            elif key == 'version':
                # NOTE(danms): Special handling of the version field, since
                # it is read_only and set in our init.
                setattr(service, base.get_attrname(key), db_service[key])
            elif key == 'uuid' and not db_service.get(key):
                # Leave uuid off the object if undefined in the database
                # so that it will be generated below.
                continue
            else:
                service[key] = db_service[key]

        service._context = context
        service.obj_reset_changes()

        return service

    def obj_load_attr(self, attrname):
        if not self._context:
            raise exception.OrphanedObjectError(method='obj_load_attr',
                                                objtype=self.obj_name())

        LOG.debug("Lazy-loading '%(attr)s' on %(name)s id %(id)s",
                  {'attr': attrname,
                   'name': self.obj_name(),
                   'id': self.id,
                   })
        if attrname != 'compute_node':
            raise exception.ObjectActionError(
                action='obj_load_attr',
                reason='attribute %s not lazy-loadable' % attrname)
        if self.binary == 'nova-compute':
            # Only n-cpu services have attached compute_node(s)
            compute_nodes = objects.ComputeNodeList.get_all_by_host(
                self._context, self.host)
        else:
            # NOTE(sbauza); Previous behaviour was raising a ServiceNotFound,
            # we keep it for backwards compatibility
            raise exception.ServiceNotFound(service_id=self.id)
        # NOTE(sbauza): Ironic deployments can have multiple nodes
        # for the same service, but for keeping same behaviour, returning only
        # the first elem of the list
        self.compute_node = compute_nodes[0]

    @base.remotable_classmethod
    def get_by_id(cls, context, service_id):
        db_service = db.service_get(context, service_id)
        return cls._from_db_object(context, cls(), db_service)

    @base.remotable_classmethod
    def get_by_uuid(cls, context, service_uuid):
        db_service = db.service_get_by_uuid(context, service_uuid)
        return cls._from_db_object(context, cls(), db_service)

    @base.remotable_classmethod
    def get_by_host_and_topic(cls, context, host, topic):
        db_service = db.service_get_by_host_and_topic(context, host, topic)
        return cls._from_db_object(context, cls(), db_service)

    @base.remotable_classmethod
    def get_by_host_and_binary(cls, context, host, binary):
        try:
            db_service = db.service_get_by_host_and_binary(context,
                                                           host, binary)
        except exception.HostBinaryNotFound:
            return
        return cls._from_db_object(context, cls(), db_service)

    @staticmethod
    @db.select_db_reader_mode
    def _db_service_get_by_compute_host(context, host, use_slave=False):
        return db.service_get_by_compute_host(context, host)

    @base.remotable_classmethod
    def get_by_compute_host(cls, context, host, use_slave=False):
        db_service = cls._db_service_get_by_compute_host(context, host,
                                                         use_slave=use_slave)
        return cls._from_db_object(context, cls(), db_service)

    # NOTE(ndipanov): This is deprecated and should be removed on the next
    # major version bump
    @base.remotable_classmethod
    def get_by_args(cls, context, host, binary):
        db_service = db.service_get_by_host_and_binary(context, host, binary)
        return cls._from_db_object(context, cls(), db_service)

    def _check_minimum_version(self):
        """Enforce that we are not older that the minimum version.

        This is a loose check to avoid creating or updating our service
        record if we would do so with a version that is older that the current
        minimum of all services. This could happen if we were started with
        older code by accident, either due to a rollback or an old and
        un-updated node suddenly coming back onto the network.

        There is technically a race here between the check and the update,
        but since the minimum version should always roll forward and never
        backwards, we don't need to worry about doing it atomically. Further,
        the consequence for getting this wrong is minor, in that we'll just
        fail to send messages that other services understand.
        """
        if not self.obj_attr_is_set('version'):
            return
        if not self.obj_attr_is_set('binary'):
            return
        minver = self.get_minimum_version(self._context, self.binary)
        if minver > self.version:
            raise exception.ServiceTooOld(thisver=self.version,
                                          minver=minver)

    @base.remotable
    def create(self):
        if self.obj_attr_is_set('id'):
            raise exception.ObjectActionError(action='create',
                                              reason='already created')
        self._check_minimum_version()
        updates = self.obj_get_changes()

        if 'uuid' not in updates:
            updates['uuid'] = uuidutils.generate_uuid()
            self.uuid = updates['uuid']

        db_service = db.service_create(self._context, updates)
        self._from_db_object(self._context, self, db_service)
        self._send_notification(fields.NotificationAction.CREATE)

    @base.remotable
    def save(self):
        updates = self.obj_get_changes()
        updates.pop('id', None)
        self._check_minimum_version()
        db_service = db.service_update(self._context, self.id, updates)
        self._from_db_object(self._context, self, db_service)

        self._send_status_update_notification(updates)

    def _send_status_update_notification(self, updates):
        # Note(gibi): We do not trigger notification on version as that field
        # is always dirty, which would cause that nova sends notification on
        # every other field change. See the comment in save() too.
        if set(updates.keys()).intersection(
                {'disabled', 'disabled_reason', 'forced_down'}):
            self._send_notification(fields.NotificationAction.UPDATE)

    def _send_notification(self, action):
        payload = service_notification.ServiceStatusPayload(self)
        service_notification.ServiceStatusNotification(
            publisher=notification.NotificationPublisher.from_service_obj(
                self),
            event_type=notification.EventType(
                object='service',
                action=action),
            priority=fields.NotificationPriority.INFO,
            payload=payload).emit(self._context)

    @base.remotable
    def destroy(self):
        db.service_destroy(self._context, self.id)
        self._send_notification(fields.NotificationAction.DELETE)

    @classmethod
    def enable_min_version_cache(cls):
        cls.clear_min_version_cache()
        cls._SERVICE_VERSION_CACHING = True

    @classmethod
    def clear_min_version_cache(cls):
        cls._MIN_VERSION_CACHE = {}

    @staticmethod
    @db.select_db_reader_mode
    def _db_service_get_minimum_version(context, binaries, use_slave=False):
        return db.service_get_minimum_version(context, binaries)

    @base.remotable_classmethod
    def get_minimum_version_multi(cls, context, binaries, use_slave=False):
        if not all(binary.startswith('nova-') for binary in binaries):
            LOG.warning('get_minimum_version called with likely-incorrect '
                        'binaries `%s\'', ','.join(binaries))
            raise exception.ObjectActionError(action='get_minimum_version',
                                              reason='Invalid binary prefix')

        if (not cls._SERVICE_VERSION_CACHING or
              any(binary not in cls._MIN_VERSION_CACHE
                  for binary in binaries)):
            min_versions = cls._db_service_get_minimum_version(
                context, binaries, use_slave=use_slave)
            if min_versions:
                min_versions = {binary: version or 0
                                for binary, version in
                                min_versions.items()}
                cls._MIN_VERSION_CACHE.update(min_versions)
        else:
            min_versions = {binary: cls._MIN_VERSION_CACHE[binary]
                            for binary in binaries}

        if min_versions:
            version = min(min_versions.values())
        else:
            version = 0
        # NOTE(danms): Since our return value is not controlled by object
        # schema, be explicit here.
        version = int(version)

        return version

    @base.remotable_classmethod
    def get_minimum_version(cls, context, binary, use_slave=False):
        return cls.get_minimum_version_multi(context, [binary],
                                             use_slave=use_slave)


def get_minimum_version_all_cells(context, binaries, require_all=False):
    """Get the minimum service version, checking all cells.

    This attempts to calculate the minimum service version for a set
    of binaries across all the cells in the system. If require_all
    is False, then any cells that fail to report a version will be
    ignored (assuming they won't be candidates for scheduling and thus
    excluding them from the minimum version calculation is reasonable).
    If require_all is True, then a failing cell will cause this to raise
    exception.CellTimeout, as would be appropriate for gating some
    data migration until everything is new enough.

    Note that services that do not report a positive version are excluded
    from this, as it crosses all cells which will naturally not have all
    services.
    """

    if not all(binary.startswith('nova-') for binary in binaries):
        LOG.warning('get_minimum_version_all_cells called with '
                    'likely-incorrect binaries `%s\'', ','.join(binaries))
        raise exception.ObjectActionError(
            action='get_minimum_version_all_cells',
            reason='Invalid binary prefix')

    # NOTE(danms): Instead of using Service.get_minimum_version_multi(), we
    # replicate the call directly to the underlying DB method here because
    # we want to defeat the caching and we need to filter non-present
    # services differently from the single-cell method.

    results = nova_context.scatter_gather_all_cells(
        context,
        Service._db_service_get_minimum_version,
        binaries)

    min_version = None
    for cell_uuid, result in results.items():
        if result is nova_context.did_not_respond_sentinel:
            LOG.warning('Cell %s did not respond when getting minimum '
                        'service version', cell_uuid)
            if require_all:
                raise exception.CellTimeout()
        elif isinstance(result, Exception):
            LOG.warning('Failed to get minimum service version for cell %s',
                        cell_uuid)
            if require_all:
                # NOTE(danms): Okay, this isn't necessarily a timeout, but
                # it's functionally the same from the caller's perspective
                # and we logged the fact that it was actually a failure
                # for the forensic investigator during the scatter/gather
                # routine.
                raise exception.CellTimeout()
        else:
            # NOTE(danms): Don't consider a zero or None result as the minimum
            # since we're crossing cells and will likely not have all the
            # services being probed.
            relevant_versions = [version for version in result.values()
                                 if version]
            if relevant_versions:
                min_version_cell = min(relevant_versions)
                min_version = (min(min_version, min_version_cell)
                               if min_version else min_version_cell)

    # NOTE(danms): If we got no matches at all (such as at first startup)
    # then report that as zero to be consistent with the other such
    # methods.
    return min_version or 0


@base.NovaObjectRegistry.register
class ServiceList(base.ObjectListBase, base.NovaObject):
    # Version 1.0: Initial version
    #              Service <= version 1.2
    # Version 1.1  Service version 1.3
    # Version 1.2: Service version 1.4
    # Version 1.3: Service version 1.5
    # Version 1.4: Service version 1.6
    # Version 1.5: Service version 1.7
    # Version 1.6: Service version 1.8
    # Version 1.7: Service version 1.9
    # Version 1.8: Service version 1.10
    # Version 1.9: Added get_by_binary() and Service version 1.11
    # Version 1.10: Service version 1.12
    # Version 1.11: Service version 1.13
    # Version 1.12: Service version 1.14
    # Version 1.13: Service version 1.15
    # Version 1.14: Service version 1.16
    # Version 1.15: Service version 1.17
    # Version 1.16: Service version 1.18
    # Version 1.17: Service version 1.19
    # Version 1.18: Added include_disabled parameter to get_by_binary()
    # Version 1.19: Added get_all_computes_by_hv_type()
    VERSION = '1.19'

    fields = {
        'objects': fields.ListOfObjectsField('Service'),
        }

    @base.remotable_classmethod
    def get_by_topic(cls, context, topic):
        db_services = db.service_get_all_by_topic(context, topic)
        return base.obj_make_list(context, cls(context), objects.Service,
                                  db_services)

    # NOTE(paul-carlton2): In v2.0 of the object the include_disabled flag
    # will be removed so both enabled and disabled hosts are returned
    @base.remotable_classmethod
    def get_by_binary(cls, context, binary, include_disabled=False):
        db_services = db.service_get_all_by_binary(
            context, binary, include_disabled=include_disabled)
        return base.obj_make_list(context, cls(context), objects.Service,
                                  db_services)

    @base.remotable_classmethod
    def get_by_host(cls, context, host):
        db_services = db.service_get_all_by_host(context, host)
        return base.obj_make_list(context, cls(context), objects.Service,
                                  db_services)

    @base.remotable_classmethod
    def get_all(cls, context, disabled=None, set_zones=False):
        db_services = db.service_get_all(context, disabled=disabled)
        if set_zones:
            db_services = availability_zones.set_availability_zones(
                context, db_services)
        return base.obj_make_list(context, cls(context), objects.Service,
                                  db_services)

    @base.remotable_classmethod
    def get_all_computes_by_hv_type(cls, context, hv_type):
        db_services = db.service_get_all_computes_by_hv_type(
            context, hv_type, include_disabled=False)
        return base.obj_make_list(context, cls(context), objects.Service,
                                  db_services)



# Copyright 2013 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Tests for keypair API."""

import mock
from oslo_concurrency import processutils
from oslo_config import cfg
from oslo_limit import fixture as limit_fixture

from nova.compute import api as compute_api
from nova import context
from nova import exception
from nova.limit import local as local_limit
from nova.objects import keypair as keypair_obj
from nova import quota
from nova.tests.unit.compute import test_compute
from nova.tests.unit import fake_crypto
from nova.tests.unit.objects import test_keypair
from nova.tests.unit import utils as test_utils

CONF = cfg.CONF


class KeypairAPITestCase(test_compute.BaseTestCase):
    def setUp(self):
        super(KeypairAPITestCase, self).setUp()
        self.keypair_api = compute_api.KeypairAPI()
        self.ctxt = context.RequestContext('fake', 'fake')
        self._keypair_db_call_stubs()
        self.existing_key_name = 'fake existing key name'
        self.pub_key = ('ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAAABAQDLnVkqJu9WVf'
                        '/5StU3JCrBR2r1s1j8K1tux+5XeSvdqaM8lMFNorzbY5iyoBbR'
                        'S56gy1jmm43QsMPJsrpfUZKcJpRENSe3OxIIwWXRoiapZe78u/'
                        'a9xKwj0avFYMcws9Rk9iAB7W4K1nEJbyCPl5lRBoyqeHBqrnnu'
                        'XWEgGxJCK0Ah6wcOzwlEiVjdf4kxzXrwPHyi7Ea1qvnNXTziF8'
                        'yYmUlH4C8UXfpTQckwSwpDyxZUc63P8q+vPbs3Q2kw+/7vvkCK'
                        'HJAXVI+oCiyMMfffoTq16M1xfV58JstgtTqAXG+ZFpicGajREU'
                        'E/E3hO5MGgcHmyzIrWHKpe1n3oEGuz')
        self.fingerprint = '4e:48:c6:a0:4a:f9:dd:b5:4c:85:54:5a:af:43:47:5a'
        self.keypair_type = keypair_obj.KEYPAIR_TYPE_SSH
        self.key_destroyed = False

    def _keypair_db_call_stubs(self):

        def db_key_pair_get_all_by_user(context, user_id, limit, marker):
            return [dict(test_keypair.fake_keypair,
                         name=self.existing_key_name,
                         public_key=self.pub_key,
                         fingerprint=self.fingerprint)]

        def db_key_pair_create(context, keypair):
            return dict(test_keypair.fake_keypair, **keypair)

        def db_key_pair_destroy(context, user_id, name):
            if name == self.existing_key_name:
                self.key_destroyed = True

        def db_key_pair_get(context, user_id, name):
            if name == self.existing_key_name and not self.key_destroyed:
                return dict(test_keypair.fake_keypair,
                            name=self.existing_key_name,
                            public_key=self.pub_key,
                            fingerprint=self.fingerprint)
            else:
                raise exception.KeypairNotFound(user_id=user_id, name=name)

        self.stub_out(
            'nova.objects.KeyPairList._get_from_db',
            db_key_pair_get_all_by_user)
        self.stub_out('nova.objects.KeyPair._create_in_db', db_key_pair_create)
        self.stub_out(
            'nova.objects.KeyPair._destroy_in_db', db_key_pair_destroy)
        self.stub_out('nova.objects.KeyPair._get_from_db', db_key_pair_get)

    def _check_notifications(self, action='create', key_name='foo'):
        self.assertEqual(2, len(self.notifier.notifications))

        n1 = self.notifier.notifications[0]
        self.assertEqual('INFO', n1.priority)
        self.assertEqual('keypair.%s.start' % action, n1.event_type)
        self.assertEqual('api.%s' % CONF.host, n1.publisher_id)
        self.assertEqual('fake', n1.payload['user_id'])
        self.assertEqual('fake', n1.payload['tenant_id'])
        self.assertEqual(key_name, n1.payload['key_name'])

        n2 = self.notifier.notifications[1]
        self.assertEqual('INFO', n2.priority)
        self.assertEqual('keypair.%s.end' % action, n2.event_type)
        self.assertEqual('api.%s' % CONF.host, n2.publisher_id)
        self.assertEqual('fake', n2.payload['user_id'])
        self.assertEqual('fake', n2.payload['tenant_id'])
        self.assertEqual(key_name, n2.payload['key_name'])


class CreateImportSharedTestMixIn(object):
    """Tests shared between create and import_key.

    Mix-in pattern is used here so that these `test_*` methods aren't picked
    up by the test runner unless they are part of a 'concrete' test case.
    """

    def assertKeypairRaises(self, exc_class, expected_message, name):
        func = getattr(self.keypair_api, self.func_name)

        args = []
        if self.func_name == 'import_key_pair':
            args.append(self.pub_key)
        args.append(self.keypair_type)

        exc = self.assertRaises(exc_class, func, self.ctxt, self.ctxt.user_id,
                                name, *args)
        self.assertIn(expected_message, str(exc))

    def assertInvalidKeypair(self, expected_message, name):
        msg = 'Keypair data is invalid: %s' % expected_message
        self.assertKeypairRaises(exception.InvalidKeypair, msg, name)

    def test_name_too_short(self):
        msg = ('Keypair name must be string and between 1 '
               'and 255 characters long')
        self.assertInvalidKeypair(msg, '')

    def test_name_too_long(self):
        msg = ('Keypair name must be string and between 1 '
               'and 255 characters long')
        self.assertInvalidKeypair(msg, 'x' * 256)

    def test_invalid_chars(self):
        msg = "Keypair name contains unsafe characters"
        self.assertInvalidKeypair(msg, '* BAD CHARACTERS!  *')

    def test_already_exists(self):
        def db_key_pair_create_duplicate(context, keypair):
            raise exception.KeyPairExists(key_name=keypair.get('name', ''))

        self.stub_out(
            'nova.objects.KeyPair._create_in_db', db_key_pair_create_duplicate)

        msg = ("Key pair '%(key_name)s' already exists." %
               {'key_name': self.existing_key_name})
        self.assertKeypairRaises(exception.KeyPairExists, msg,
                                 self.existing_key_name)

    @mock.patch.object(quota.QUOTAS, 'count_as_dict',
                       return_value={'user': {
                                         'key_pairs': CONF.quota.key_pairs}})
    def test_quota_limit(self, mock_count_as_dict):
        msg = "Quota exceeded, too many key pairs."
        self.assertKeypairRaises(exception.KeypairLimitExceeded, msg, 'foo')

    def _test_quota_during_recheck(self, mock_method, msg):
        # Skip for import key pair due to bug 1959732.
        if self.func_name == 'import_key_pair':
            self.skipTest('bug/1959732: import_key_pair missing quota recheck')

        self.assertKeypairRaises(exception.KeypairLimitExceeded, msg, 'foo')
        self.assertEqual(2, mock_method.call_count)

    @mock.patch('nova.objects.Quotas.check_deltas')
    def test_quota_during_recheck(self, mock_check):
        """Simulate a race where this request initially has enough quota to
        progress partially through the create path but then fails the quota
        recheck because a parallel request filled up the quota first.
        """
        # First quota check succeeds, second (recheck) fails.
        mock_check.side_effect = [None,
                                  exception.OverQuota(overs='key_pairs')]
        msg = "Quota exceeded, too many key pairs."
        self._test_quota_during_recheck(mock_check, msg)

    def test_quota_unified_limits(self):
        self.flags(driver="nova.quota.UnifiedLimitsDriver", group="quota")
        self.useFixture(limit_fixture.LimitFixture(
            {'server_key_pairs': 0}, {}))
        msg = ("Resource %s is over limit" % local_limit.KEY_PAIRS)
        self.assertKeypairRaises(exception.KeypairLimitExceeded, msg, 'foo')

    @mock.patch('nova.limit.local.enforce_db_limit')
    def test_quota_during_recheck_unified_limits(self, mock_enforce):
        """Simulate a race where this request initially has enough quota to
        progress partially through the create path but then fails the quota
        recheck because a parallel request filled up the quota first.
        """
        self.flags(driver="nova.quota.UnifiedLimitsDriver", group="quota")
        self.useFixture(limit_fixture.LimitFixture(
            {'server_key_pairs': 100}, {}))
        # First quota check succeeds, second (recheck) fails.
        mock_enforce.side_effect = [
            None, exception.KeypairLimitExceeded('oslo.limit message')]
        msg = 'oslo.limit message'
        self._test_quota_during_recheck(mock_enforce, msg)


class CreateKeypairTestCase(KeypairAPITestCase, CreateImportSharedTestMixIn):
    func_name = 'create_key_pair'

    @mock.patch('nova.compute.utils.notify_about_keypair_action')
    def _check_success(self, mock_notify):
        keypair, private_key = self.keypair_api.create_key_pair(
            self.ctxt, self.ctxt.user_id, 'foo', key_type=self.keypair_type)
        self.assertEqual('foo', keypair['name'])
        self.assertEqual(self.keypair_type, keypair['type'])
        mock_notify.assert_has_calls([
            mock.call(context=self.ctxt, keypair=keypair,
                      action='create', phase='start'),
            mock.call(context=self.ctxt, keypair=keypair,
                      action='create', phase='end')])
        self._check_notifications()

    def test_success_ssh(self):
        self._check_success()

    def test_success_x509(self):
        self.keypair_type = keypair_obj.KEYPAIR_TYPE_X509
        self._check_success()

    def test_x509_subject_too_long(self):
        # X509 keypairs will fail if the Subject they're created with
        # is longer than 64 characters. The previous unit tests could not
        # detect the issue because the ctxt.user_id was too short.
        # This unit tests is added to prove this issue.
        self.keypair_type = keypair_obj.KEYPAIR_TYPE_X509
        self.ctxt.user_id = 'a' * 65
        self.assertRaises(processutils.ProcessExecutionError,
                          self._check_success)

    def test_success_unified_limits(self):
        self.flags(driver="nova.quota.UnifiedLimitsDriver", group="quota")
        self.useFixture(limit_fixture.LimitFixture(
            {'server_key_pairs': 1}, {}))
        self._check_success()

    @mock.patch('nova.objects.Quotas.check_deltas')
    def test_quota_recheck_disabled(self, mock_check):
        self.flags(recheck_quota=False, group="quota")
        self._check_success()
        self.assertEqual(1, mock_check.call_count)

    @mock.patch('nova.limit.local.enforce_db_limit')
    def test_quota_recheck_disabled_unified_limits(self, mock_enforce):
        self.flags(driver="nova.quota.UnifiedLimitsDriver", group="quota")
        self.flags(recheck_quota=False, group="quota")
        self.useFixture(limit_fixture.LimitFixture(
            {'server_key_pairs': 1}, {}))
        self._check_success()
        self.assertEqual(1, mock_enforce.call_count)


class ImportKeypairTestCase(KeypairAPITestCase, CreateImportSharedTestMixIn):
    func_name = 'import_key_pair'

    @mock.patch('nova.compute.utils.notify_about_keypair_action')
    def _check_success(self, mock_notify):
        keypair = self.keypair_api.import_key_pair(self.ctxt,
                                                   self.ctxt.user_id,
                                                   'foo',
                                                   self.pub_key,
                                                   self.keypair_type)

        self.assertEqual('foo', keypair['name'])
        self.assertEqual(self.keypair_type, keypair['type'])
        self.assertEqual(self.fingerprint, keypair['fingerprint'])
        self.assertEqual(self.pub_key, keypair['public_key'])
        self.assertEqual(self.keypair_type, keypair['type'])
        mock_notify.assert_has_calls([
            mock.call(context=self.ctxt, keypair=keypair,
                      action='import', phase='start'),
            mock.call(context=self.ctxt, keypair=keypair,
                      action='import', phase='end')])
        self._check_notifications(action='import')

    def test_success_ssh(self):
        self._check_success()

    def test_success_ssh_ed25519(self):
        self.pub_key = ('ssh-ed25519 AAAAC3NzaC1lZDI1NTE5AAAAIFl6u75BTi8xGtSPm'
                        '1yVJuLE/oMtCOuEMJJnBSuZEdXz')
        self.fingerprint = '1a:1d:a7:2c:4c:ff:15:c4:70:13:38:b6:ac:4c:dc:12'
        self._check_success()

    def test_success_x509(self):
        self.keypair_type = keypair_obj.KEYPAIR_TYPE_X509
        certif, fingerprint = fake_crypto.get_x509_cert_and_fingerprint()
        self.pub_key = certif
        self.fingerprint = fingerprint
        self._check_success()

    def test_bad_key_data(self):
        exc = self.assertRaises(exception.InvalidKeypair,
                                self.keypair_api.import_key_pair,
                                self.ctxt, self.ctxt.user_id, 'foo',
                                'bad key data')
        msg = u'Keypair data is invalid: failed to generate fingerprint'
        self.assertEqual(msg, str(exc))

    def test_success_unified_limits(self):
        self.flags(driver="nova.quota.UnifiedLimitsDriver", group="quota")
        self.useFixture(limit_fixture.LimitFixture(
            {'server_key_pairs': 1}, {}))
        self._check_success()

    @mock.patch('nova.objects.Quotas.check_deltas')
    def test_quota_recheck_disabled(self, mock_check):
        self.flags(recheck_quota=False, group="quota")
        self._check_success()
        self.assertEqual(1, mock_check.call_count)

    @mock.patch('nova.limit.local.enforce_db_limit')
    def test_quota_recheck_disabled_unified_limits(self, mock_enforce):
        self.flags(driver="nova.quota.UnifiedLimitsDriver", group="quota")
        self.flags(recheck_quota=False, group="quota")
        self.useFixture(limit_fixture.LimitFixture(
            {'server_key_pairs': 1}, {}))
        self._check_success()
        self.assertEqual(1, mock_enforce.call_count)


class GetKeypairTestCase(KeypairAPITestCase):
    def test_success(self):
        keypair = self.keypair_api.get_key_pair(self.ctxt,
                                                self.ctxt.user_id,
                                                self.existing_key_name)
        self.assertEqual(self.existing_key_name, keypair['name'])


class GetKeypairsTestCase(KeypairAPITestCase):
    def test_success(self):
        keypairs = self.keypair_api.get_key_pairs(self.ctxt, self.ctxt.user_id)
        self.assertEqual([self.existing_key_name],
                         [k['name'] for k in keypairs])


class DeleteKeypairTestCase(KeypairAPITestCase):
    @mock.patch('nova.compute.utils.notify_about_keypair_action')
    def test_success(self, mock_notify):
        self.keypair_api.delete_key_pair(self.ctxt, self.ctxt.user_id,
                self.existing_key_name)
        self.assertRaises(exception.KeypairNotFound,
                self.keypair_api.get_key_pair, self.ctxt, self.ctxt.user_id,
                self.existing_key_name)

        match_by_name = test_utils.CustomMockCallMatcher(
            lambda keypair: keypair['name'] == self.existing_key_name)

        mock_notify.assert_has_calls([
            mock.call(context=self.ctxt,
                      keypair=match_by_name,
                      action='delete', phase='start'),
            mock.call(context=self.ctxt,
                      keypair=match_by_name,
                      action='delete', phase='end')])
        self._check_notifications(action='delete',
                key_name=self.existing_key_name)
